// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef ORDERINGENTITY_HPP
#define ORDERINGENTITY_HPP

#include <map>
#include "GridPoint.hpp"
#include "ImageListTypes.hpp"
#include "GetIlsi.hpp"

class OrderingHeuristic;
class PerformanceMeter;

/** An OrderingEntity and an OrderingHeuristic are tightly coupled.  In many
 ** cases there is a one-to-one mapping between the two.  In these cases,
 ** things would be simpler if the two classes were collapsed into one.  But
 ** OrderingHeuristics S and E (and any other non-fixedOrder heuristics that
 ** might be defined in the future) need the same image general purpose lists
 ** that make no assumptions about the order in which grid points are selected
 ** and so perform no POF filtering on the lists.  You definitely don't want
 ** to create these general lists twice, as they are the most expensive lists
 ** both a memory and processing perspective.  For this reason the image lists
 ** defined by an OrderingHeuristic and all the machinery dedicated to
 ** creating and accessing those lists, is relegated to this separate (and
 ** potentially shared) object.
 **/
class OrderingEntity
{
    public:
        OrderingEntity(
                OrderingHeuristic*        oh,
                const DlxHead*            dlxRoot,
                int                       numShapes,
                const int*                remainingShapeList,
                GridPoint const * const * filaGrid,
                bool                      nof,
                bool                      need6Bit);

        virtual ~OrderingEntity();

        ImageListSet getImageListSet(const GridPoint* gp, const gridbitfield_t occupancy) const;

        void loadImages(
                int nrp,
                int numFilaGridPoints,
                const gridbitfield_t neighborOccupancy[][64]);

        void clearImages();

        /** Pointer to the first OrderingHeuristic that requests this
         ** OrderingEntity.  Currently this reference is only used by the
         ** entity to invoke the heuristic's initSelectionOrder() method which
         ** sets up the selectionOrder array for fixedOrder heuristics.
         **/
        OrderingHeuristic* oh;

        // The following are all set at construction time and are all just
        // copies of fixed setttings, or const pointers to dynamic
        // information maintained by other entities, that is often needed by
        // OrderingEntities/OrderingHeuristics.
        //
        const DlxHead*            dlxRoot;
        int                       numShapes;
        const int*                remainingShapeList;
        GridPoint const * const * filaGrid;
        bool                      nof;

        /** Lower bound for an ILSI index in for loops.
         **
         ** If NOF is disabled then ILSI is not used, and ilsi is just
         ** replaced with 0 everywhere in the code, and minIlsi is likewise
         ** set to 0.  For the commonly used "f" heuristic, only ilis
         ** from 56 to 63 can possibly occur, and so minIlsi is set to
         ** 56 to reduce wasted memory allocations and eliminate loops
         ** over the unused ILSI range from 0 to 55.  Otherwise
         ** this is some other hueristic and NOF is enabled so the
         ** full range of ILSI 0 to 63 is available, and again minIlsi
         ** takes on the value 0.
         **/
        int                       minIlsi;

        /** Upper bound for an ILSI index in for loops.
         **
         ** If NOF is disabled then ILSI is not used, and ilsi is just
         ** replaced with 0 everywhere in the code, and maxIlsi is set
         ** to 1.  Otherwise, when NOF is disabled, the maximum
         ** ILSI is always 63 and so maxIlsi is set to 64.
         **/
        int                       maxIlsi;

        /** The number of FILA grid points.  Set each time the image lists
         ** are reinitialized via a load() call.
         **/
        int                   numFilaGridPoints;

        // The next three fields (selectionOrder, selectionOrderMap, and
        // lastSelectionIndex) are only used by fixedOrder heuristics.

        /** A sorted array of GridPoint->bitNum defining the order that
         ** GridPoints are selected by the related OrderingHeuristic.  This
         ** is never initialized (and never used) by the general ("G")
         ** OrderingEntity since there is no predefined selection order
         ** for that entity.
         **/
        int selectionOrder[maxFilaGridPoints];

        /** For fixedOrder heuristics, when selectFila is invoked with
         ** numRemainingPieces = some number k, the order index of the
         ** selected GridPoint is stored in lastSelectionIndex[k].  Then
         ** for a subsequent recursive call (with numRemainingPieces = k-1),
         ** the forward search through the selectionOrder array for an
         ** unoccupied GridPoint may begin at lastSelectionIndex[k]+1.
         **
         ** Note:  this array is oversized.  The 0th entry is never used
         **        since you never call select when there are 0 pieces left
         **        (the puzzle is solved at that point obviously).  Positions
         **        1 through numRemainingPieces+1 (inclusive) are used.
         **        Normally numRemainingPieces is always the same for each
         **        initFila call (which can happen multiple times during one
         **        solve), but because there's always a maximum volume limit,
         **        the initFila call can be delayed if pieces have different
         **        volumes and the fila activation setting is large enough
         **        that the volume constraint starts to kick in.  (I should
         **        really change activation thresholds to be volume based.)
         **        Every piece must have size at least one, so the number
         **        numRemainingGridPoints+2 is guaranteed to be sufficiently
         **        large.  The oversizing is inconsequential as there are only
         **        ever a few (say 1 to 4) OrderingEntities, and the unused
         **        array entries aren't even initialized.
         **
         ** Note:  For all entities, the array should be initially filled with
         **        the value -1 to ensure that the first time the entity is
         **        used, the search starts at position zero.
         **
         ** Note:  If this entity is for any reason ever used at position k
         **        then inconsistently skipped at position k-1, then used
         **        again at position k-2, you'll need some way to copy the
         **        value at k to k-1.  I was thinking this might happen
         **        if I eventually implement OH priority override.  This
         **        planned feature would support, for example, an S or E
         **        heuristic to first be consulted to search for a hole with a
         **        very limited number of fit possibilities (say only 0 or 1)
         **        and, if one is found, it is forcibly selected, but
         **        otherwise, some fixedOrder heuristic (say F) is invoked to
         **        achieve a more ordered fill behavior.  In that event it
         **        should be easy and efficient to invoke some method on
         **        the skipped heuristic to do this data promotion.
         **/
        int lastSelectionIndex[maxFilaGridPoints + 2];

        PerformanceMeter* filaLoadMeter;

    private:

        void initPriorityOccupancy(gridbitfield_t* priorityOccupancy);

    protected:

        /** getIlsi[g] is the a pointer to the most trimmed down getIlsi
         ** function that can safely be used for GridPoint g.  A
         ** general-purpose getIlsi function will examine the occupancy state
         ** of all six neighbors of g and convert the holes found to an
         ** index number (from 0 to 63) which may be used to index
         ** to the appropriate ImageListSet which is guaranteed to have
         ** no images that conflict with neighbors of g that are occupied.
         ** But it's sorta stupid, for example, to check the occupancy state for
         ** the neighbors in the +z and -z directions for a gridpoint in a
         ** 2-D puzzle.  Similarly, the grid point in the lower left corner
         ** need not check occupancy for neighbors below and to the left
         ** of that GridPoint (since again their are none).  Furthermore,
         ** some ordering heuristics guarantee a fill order which may
         ** guarantee that additional neighbors of g are occupied.  You don't
         ** wanna be checking the occupancy state of these neighbors over and
         ** over everytime you try to fill g.  getIlsi[g] will only check
         ** occupancy states of neighbors that actually exist, and could
         ** possibly be open.
         **
         ** So the function itself is faster, but there is some overhead to
         ** this array lookup and the associated function call (since it can't
         ** be inlined).  Compared to the approach of always checking all 6
         ** neighbors, I've found this approach to speed things (maybe 5-10%)
         ** for 2D puzzles using the First (f) heuristic.  But it slows things
         ** down (maybe 3-4%) for 3D puzzles using the S or e heuristics
         ** (which is really annoying).  I've tried various ways of
         ** implementing this but was never able to reduce this overhead.
         ** What I really want is an indexed jump (goto) command.
         **/
        GetIlsi getIlsi[maxFilaGridPoints];

        /** Basically a 4 dimensional array of Image pointers.  The the inner
         ** index currently actually acts on a vector but I may one day change
         ** that to an array as well to (possibly) improve performance and
         ** (definitely) save memory.  The array is indexed like this:
         **
         **   this->all[gp->bitNum][ili][shape->id][i]
         **
         ** where
         **     gp  is a GridPoint;
         **     ili is an image list index;
         **     shape is just an available shape; and
         **     i is an index into the vector of image pointers.
         **
         ** selectFila returns a two dimensional subarray of this beast given
         ** by all[gp->bitNum][ili] where gp is the selected GridPoint and ili
         ** is the appropriate nof image list index (or just zero if NOF is
         ** disabled).
         **/
        ImageListSet** all;
};


inline ImageListSet OrderingEntity::getImageListSet(const GridPoint* gp, const gridbitfield_t occupancy) const
{
    if(nof)
        return all[gp->bitNum][(*getIlsi[gp->bitNum])(gp, occupancy)];
    else
        return all[gp->bitNum][0];
}

#endif
