// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef ORDERINGHEURISTICSTORE_HPP
#define ORDERINGHEURISTICSTORE_HPP

#include <map>
#include <set>

#include "GridPoint.hpp"
#include "ImageListTypes.hpp"
#include "OrderingHeuristic.hpp"

class OrderingEntity;

typedef std::map<std::string, OrderingEntity*> EntityMap;
typedef std::set<OrderingEntity*>              EntitySet;

class OrderingHeuristicStore
{
    public:
        OrderingHeuristicStore();
        virtual ~OrderingHeuristicStore();

        void setDefaultXc(double defaultXc);
        void setDefaultYc(double defaultYc);
        void setDefaultZc(double defaultZc);
        void setDlxRoot(const DlxHead* dlxRoot);
        void setNumShapes(int numShapes);
        void setNumMobilePieces(int numMobilePieces);
        void setRemainingShapeList(const int* remainingShapeList);
        void setFila(int fila);
        void setFilaGrid(GridPoint const * const * filaGrid);
        void setNof(bool nof);
        void setHeuristic(int nrp, const std::string& def);
        void loadImages(int nrp, int numFilaGridPoints);
        DlxHead* selectDlx(int nrp) const;
        ImageListSet selectFila(int nrp, int nrs, const gridbitfield_t occupancy, bool dlxPristine = false) const;
        void clearImages();

    private:
        void storeHeuristic(int nrp, OrderingHeuristic* oh);

        /** Only instantiates an OrderingEntity of the given name if it does
         ** not already exist.  Also stores a reference to the
         ** OrderingHeuristic that first requested it.  This is only needed
         ** for grid-point ordering which in turn is used to initialize the
         ** priorityOccupancy bit fields during each load() call (below).  If
         ** two heuristics request the same resource, they must have the same
         ** ordering requirements, hence only one reference is needed.
         **/
        OrderingEntity* genEntity(OrderingHeuristic* oh);

        void initNeighborOccupancy(gridbitfield_t neighborOccupancy[][64]);


        // Default puzzle center for radial and angular heuristics.
        //
        double defaultXc;
        double defaultYc;
        double defaultZc;

        const DlxHead*            dlxRoot;
        int                       numShapes;
        int                       numMobilePieces;
        const int*                remainingShapeList;
        int                       fila;
        GridPoint const * const * filaGrid;
        bool                      nof;
        int                       numFilaGridPoints;

        /** Array of heuristic pointers indexed by number of remainingPieces.
         **/
        OrderingHeuristic**       heuristicList;

        /** Map of OrderingEntity.  There's one entry for each uniquely
         ** named entity requested by the set of OrderingHeuristics
         ** configured for this run of the solver.
         **/
        EntityMap                 entityMap;

        EntitySet                 filaEntitySet;

        /** If NOF is enabled, the question arises as to whether a 6-bit
         ** (64-entry) nieghbor-occupancy bit array is required or only a
         ** 3-bit (8-entry).  You only need 3-bit if the only Heuristic used
         ** during the FILA portion of the search, but 6-bit is needed by
         ** every other heuristic during that phase of the search.  (I.e., for
         ** a numRemainingPieces setting at-or-below the fila setting.)  This
         ** boolean is set if one of those other heurstics is used.  If need6Bit
         ** is false, then only 8 bit masks are calculated for each
         ** neighbor-occupancy bit mask.
         **/
        bool                      need6Bit;

        friend class OrderingHeuristic;
};



inline OrderingHeuristicStore::OrderingHeuristicStore()
    :   
        defaultXc(0),
        defaultYc(0),
        defaultZc(0),
        dlxRoot(NULL),
        numShapes(0),
        numMobilePieces(0),
        remainingShapeList(NULL),
        fila(0),
        filaGrid(NULL),
        nof(false),
        numFilaGridPoints(0),
        heuristicList(NULL),
        need6Bit(false)
{
}

inline void OrderingHeuristicStore::setDefaultXc(double defaultXc)
{
    this->defaultXc = defaultXc;
}

inline void OrderingHeuristicStore::setDefaultYc(double defaultYc)
{
    this->defaultYc = defaultYc;
}

inline void OrderingHeuristicStore::setDefaultZc(double defaultZc)
{
    this->defaultZc = defaultZc;
}

inline void OrderingHeuristicStore::setDlxRoot(const DlxHead* dlxRoot)
{
    this->dlxRoot = dlxRoot;
}

inline void OrderingHeuristicStore::setNumShapes(int numShapes)
{
    this->numShapes = numShapes;
}

inline void OrderingHeuristicStore::setRemainingShapeList(const int* remainingShapeList)
{
    this->remainingShapeList = remainingShapeList;
}

inline void OrderingHeuristicStore::setFila(int fila)
{
    this->fila = fila;
}

inline void OrderingHeuristicStore::setFilaGrid(GridPoint const * const * filaGrid)
{
    this->filaGrid = filaGrid;
}

inline void OrderingHeuristicStore::setNof(bool nof)
{
    this->nof = nof;
}

inline DlxHead* OrderingHeuristicStore::selectDlx(int nrp) const
{
    return heuristicList[nrp]->selectDlx();
}

inline ImageListSet OrderingHeuristicStore::selectFila(int nrp, int nrs, const gridbitfield_t occupancy, bool dlxPristine) const
{
    return heuristicList[nrp]->selectFila(nrp, nrs, occupancy, dlxPristine);
}

#endif
